unit TC_Class;

{*********************************************}
{                                             }
{   THERMOCHEMICAL CLASSES (Species)          }
{                                             }
{   (C) 2000 Belov G.V., IVTAN, Thermocenter  }
{                                             }
{   Please send bugs, comments and            }
{   suggestions to:                            }
{                                             }
{   gbelov@yandex.ru                           }
{                                             }
{*********************************************}

interface
uses Classes, SysUtils, Dialogs, AtConst;

const MaxNumberOfElements = 5; // max number of elements in chemical formula
const MaxNumberOfCoefficients = 7;

Type ST3 = string[3];
     ST4 = string[4];
     ST8 = string[8];
     ST10 = string[10];
     ST15 = string[15];
     ST20 = string[20];
     ST30 = string[30+15+1];
     ST50 = string[50];
     TNos = ShortInt;
     TQty = double;
     TEL_Nos = array[1..MaxNumberOfElements] of TNos;
     TEL_Qty = array[1..MaxNumberOfElements] of TQty;

Type TCoeff = array[1..MaxNumberOfCoefficients] of single;

Type TPolynom = class
       private
         FTmin  : single;     // Tmin for the temperature range
         FTmax  : single;     // Tmax for the temperature range
         FF     : TCoeff;     // coefficients of approximating function
         FPhase : ST10;       // supplementary information on phase state
       protected
         procedure SetFTmin(S : single); virtual;
         procedure SetFTmax(S : single); virtual;
         procedure SetFPhase(S : ST10); virtual;
         function GetF(i : integer) : single; virtual;
         procedure SetF(i: integer; S : single);
       public
         procedure SetAllData(Tmin, Tmax: single; F : TCoeff; Phase : ST10);
         property Tmin : single read FTmin write SetFTmin;
         property Tmax : single read FTmax write SetFTmax;
         property F[index : integer] : single read GetF write SetF;
         property Phase : ST10 read FPhase write SetFPhase;
       end;

Type TPolynoms = class
  private
    FP : TList;     // list of objects of the type TPolynom
    FTmin: single;
    FTmax: single;
    function GetPolyn(i : integer) : TPolynom;
    function FPCount: integer;
  public
    Constructor Create;
    Destructor Destroy; override;
    procedure AddPol(Tmin,Tmax: single; const F: array of single);
    procedure Clear;
    property Pol[i: integer]: TPolynom read GetPolyn;
    property Count : integer read FPCount;
    property Tmin: single read FTmin write FTmin;
    property Tmax: single read FTmax write FTmax;
  end;

Type TSpecies = class
  private
    FFormula : ST20;  // chemical formula
    FState : ST20;    // phase state(s)
    FPrecClass : ST4;  // precision class
    FName : ST50;   // name of substance
    FMolMass : double;   // molecular mass
    FTref : single;      // reference temperature
    FReaction : ST50;    // reaction of dissociation (sublimation)
    FDHR : single;   // enthalpy change of the reaction of dissociation
    FDH0 : single;   // enthalpy of formation at 0 K
    FDH298 : single;  // enthalpy of formation at 298 K
    FCp298 : single;  // heat capacity at 298 K
    FS298 : single;   // entropy at 298 K
    FH298H0 : single;  // enthalpy change H(298) - H(0)
    FSnucl : single;   // nuclear spin component
    FPrefUnit : ST3;   // units of reference pressure (atm, bar)
    FDate : ST8;    // date of the information
    FNelts : byte;  // number of elements in molecula
    FNpols : byte;   // number of polynoms
    FAddrPol : integer;  // address of the first polynom
    FPstate : byte;    // code of the phase state
    FEL_Nos : TEL_Nos;  // numbers of elements in Mendeleev Table
    FEL_Qty : TEL_Qty;  // amounts of elements in molecula
    FIonCharge : ShortInt;   // charge value
    FPolynoms : TPolynoms;  // list of polynoms
    FisomericForm: st10;    // isomeric form
    FModification: st15;
  protected
    procedure SetFormula(S : ST20); virtual;
    procedure SetState(S : ST20); virtual;
    procedure SetPrecClass(S : ST4); virtual;
    procedure SetName(S : ST50); virtual;
    procedure SetMolMass(S : double); virtual;
    procedure SetTref(S : single); virtual;
    procedure SetReaction(S : ST50); virtual;
    procedure SetDHR(S : single); virtual;
    procedure SetDH0(S : single); virtual;
    procedure SetDH298(S : single); virtual;
    procedure SetCp298(S : single); virtual;
    procedure SetS298(S : single); virtual;
    procedure SetH298H0(S : single); virtual;
    procedure SetSnucl(S : single); virtual;
    procedure SetPrefUnit(S : ST3); virtual;
    procedure SetDate(S : ST8); virtual;
    procedure SetNelts(S : byte); virtual;
    procedure SetNpols(S : byte); virtual;
    procedure SetAddrPol(S : integer); virtual;
    procedure SetPstate(S : byte); virtual;
    procedure SetEL_Nos(i: integer; S : TNos); virtual;
    procedure SetEL_Qty(i: integer; S : TQty); virtual;
    function GetFEl_Nos(i: integer): TNos; virtual;
    function GetFEl_Qty(i: integer): TQty; virtual;
    procedure SetPolynoms(S: TPolynoms); virtual;
  public
    constructor Create;
    destructor Destroy; override;
    procedure SetAllData(Formula : ST20; State : ST20;  PrecClass : ST4;
                       Name : ST50; MolMass : double; Tref : single; Reaction : ST50;
                       DHR : single; DH0 : single;  DH298 : single;
                       Cp298 : single;  S298 : single;  H298H0 : single;
                       Snucl : single; PrefUnit : ST3;  Date : ST8;
                       Nelts : byte;  Npols : byte;  AddrPol : integer;
                       Pstate : byte;  EL_Nos : TEL_Nos; EL_Qty : TEL_Qty;
                       IonCharge : ShortInt);
    procedure Clear;
    function Cp(T: single) : single; virtual;
    function S(T: single) : single;  virtual;
    function Hfull(T: single) : single; virtual;
    function H(T: single) : single; virtual;
    function G(T: single) : single; virtual;
    function F(T: single) : single; virtual;
    procedure AddExPolynoms;        virtual; // extrapolate TD properties, Cp = const
    function CalcMolMass : double;
    property Formula: ST20 read FFormula write SetFormula;
    property State: ST20 read FState write SetState;
    property PrecClass: ST4 read FPrecClass write SetPrecClass;
    property Name: ST50 read FName write SetName;
    property MolMass: double read FMolMass write SetMolMass;
    property Tref: single read FTref write SetTref;
    property Reaction: ST50 read FReaction write SetReaction;
    property DHR: single read FDHR write SetDHR;
    property DH0: single read FDH0 write SetDH0;
    property DH298: single read FDH298 write SetDH298;
    property Cp298: single read FCp298 write SetCp298;
    property S298: single read FS298 write SetS298;
    property H298H0: single read FH298H0 write SetH298H0;
    property Snucl: single read FSnucl write SetSnucl;
    property PrefUnit: ST3 read FPrefUnit write SetPrefUnit;
    property Date: ST8 read FDate write SetDate;
    property Nelts: byte read FNelts write SetNelts;
    property Npols: byte read FNpols write SetNpols;
    property AddrPol: integer read FAddrPol write SetAddrPol;
    property Pstate: byte read FPstate write SetPstate;
    property EL_Nos[index: integer]: TNos read GetFEL_Nos write SetEL_Nos;
    property EL_Qty[index: integer]: TQty read GetFEL_Qty write SetEL_Qty;
    property IonCharge : ShortInt read FIonCharge write FIonCharge;
    property Polynoms : TPolynoms read FPolynoms write SetPolynoms;
    property isomericForm: st10 read FisomericForm write FisomericForm;
    property Modification: st15 read FModification write FModification;
  end;

Type TOneSpecies = class
     private
       FFormula : ST20;       // substance formula
       Findex : integer;      // index of substance in the list
       FNos: TEL_Nos;         // numbers of chemical elements
     protected
       procedure SetFormula(S: ST20);
       procedure SetIndex(S: integer);
       procedure SetNos(S: TEL_Nos);
     public
       property Formula : ST20 read FFormula write SetFormula;
       property ix : integer read Findex write SetIndex;
       property Nos: TEL_Nos read FNos write SetNos;
     end;

Type TListOfSpecies = class
     private
       FList : TStringList;  // list of substances
     public
       constructor Create; virtual;
       destructor Destroy; override;
       procedure AddSubstance(S: TOneSpecies);
       procedure Clear;
       property List: TStringList read FList;
       function SubstanceByName(pIndex: ST20): TOneSpecies; virtual;
       function SubstanceByIndex(pIndex: integer): TOneSpecies; virtual;
     end;

Type TTPXListOfSpecies = class
     private
       FList : TStringList;  // list of substances
       function GetFCount : integer;
     public
       constructor Create; virtual;
       destructor Destroy; override;
       procedure AddSubstance(TPX_KOD: st30; S: TOneSpecies);
       procedure Clear;
       property Count: integer read GetFCount;
       function SubstanceByName(pIndex: ST20): TOneSpecies; virtual;
       function SubstanceByIndex(pIndex: integer): TOneSpecies; virtual;
     end;

Type TFullList = class
     private
       FList : TStringList;  // list of substances
       function GetFCount : integer;
     public
       constructor Create; virtual;
       destructor Destroy; override;
       procedure Clear;
       procedure AddSubstance(S: TSpecies);
       property Count: integer read GetFCount;
       function SubstanceByName(pIndex: ST20): TSpecies; virtual;
       function SubstanceByIndex(pIndex: integer): TSpecies; virtual;
     end;

  var Species : TSpecies;
      Polynom  : TPolynom;
      Polynoms : TPolynoms;
      OneSpecies : TOneSpecies;
      ListOfSpecies : TListOfSpecies;
      TPXListOfSpecies : TTPXListOfSpecies;
      FullList :TFullList;

implementation

procedure TPolynom.SetAllData(Tmin, Tmax: single; F : TCoeff; Phase : ST10);
begin
  FTmin := Tmin;
  FTmax := Tmax;
  FF := F;
  FPhase := Phase;
end;

procedure TPolynom.SetFTmin(S : single);
begin
  FTmin := S;
end;

procedure TPolynom.SetFTmax(S : single);
begin
  FTmax := S;
end;

procedure TPolynom.SetF(i : integer; S : single);
begin
  if (i > High (FF)) or (i < Low(FF)) then
    raise Exception.Create('F index out of range: '+inttostr(i));
  FF[i] := S;
end;

function TPolynom.GetF(i : integer) : single;
begin
  if (i > High (FF)) or (i < Low(FF)) then
    raise Exception.Create('F index out of range: '+inttostr(i));
  Result := FF[i];
end;

procedure TPolynom.SetFPhase(S : ST10);
begin
  FPhase := S;
end;

constructor TPolynoms.Create;
begin
  inherited Create;
  FP := TList.Create;
  FTmin := 1.e30;
end;

destructor TPolynoms.Destroy;
var i : integer;
begin
  for i := 0 to FP.Count-1 do
    begin
    Polynom := FP.Items[i];
    Polynom.Free;
    end;
  FP.Free;
  inherited Destroy;
end;

function TPolynoms.GetPolyn(i : integer) : TPolynom;
begin
  if (i < 0) or (i > FP.Count-1) then
    raise Exception.Create('Invalid index: '+IntToStr(i));
  Result := FP[i];
end;

function TPolynoms.FPCount : integer;
begin
  Result := FP.Count;
end;

procedure TPolynoms.AddPol(Tmin,Tmax: single; const F: array of single);
var Polynom : TPolynom;
    j : integer;
begin
   Polynom := TPolynom.Create;
   Polynom.Tmin := Tmin;
   Polynom.Tmax := Tmax;
   for j := Low(F) to High(F) do
     if j <= MaxNumberOfCoefficients then
       Polynom.F[j+1] := F[j];
  FP.Add(Polynom);
  if Tmin < FTmin then FTmin := Tmin;
  if Tmax > FTmax then FTmax := Tmax;
end;

procedure TPolynoms.Clear;
var i : integer;
begin
  for i := 0 to FP.Count-1 do
    begin
    Polynom := FP.Items[i];
    Polynom.Free;
    end;
  FP.Clear;
end;

constructor TSpecies.Create;
begin
  inherited Create;
  FPolynoms := TPolynoms.Create;
end;

destructor TSpecies.Destroy;
begin
  if Assigned(FPolynoms) then FPolynoms.Free;
  inherited Destroy;
end;

procedure TSpecies.SetAllData(
                       Formula : ST20; State : ST20;  PrecClass : ST4;
                       Name : ST50; MolMass: double; Tref : single; Reaction : ST50;
                       DHR : single; DH0 : single;  DH298 : single;
                       Cp298 : single;  S298 : single;  H298H0 : single;
                       Snucl : single; PrefUnit : ST3;  Date : ST8;
                       Nelts : byte;  Npols : byte;  AddrPol : integer;
                       Pstate : byte;  EL_Nos : TEL_Nos; EL_Qty : TEL_Qty;
                       IonCharge : ShortInt);
begin
    FFormula := Formula;
    FState := State;
    FPrecClass := PrecClass;
    FName := Name;
    FMolMass := MolMass;
    FTref := Tref;
    FReaction := Reaction;
    FDHR := DHR;
    FDH0 := DH0;
    FDH298 := DH298;
    FCp298 := Cp298;
    FS298 := S298;
    FH298H0 := H298H0;
    FSnucl := Snucl;
    FPrefUnit := PrefUnit;
    FDate := Date;
    FNelts := Nelts;
    FNpols := Npols;
    FAddrPol := AddrPol;
    FPstate := Pstate;
    FEL_Nos := EL_Nos;
    FEL_Qty := EL_Qty;
    FIonCharge := IonCharge;
end;

procedure TSpecies.Clear;
var i : integer;
begin
    if Assigned(Polynoms) then
      Polynoms.Clear;
    FFormula := '';
    FState := '';
    FPrecClass := '';
    FName := '';
    FMolMass := 0;
    Tref := 0;
    FReaction := '';
    FDHR := 0;
    FDH0 := 0;
    FDH298 := 0;
    FCp298 := 0;
    FS298 := 0;
    FH298H0 := 0;
    FSnucl := 0;
    FPrefUnit := '';
    FDate := '';
    FNelts := 0;
    FNpols := 0;
    FAddrPol := 0;
    FPstate := 0;
    FIonCharge := 0;
    for i := 1 to MaxNumberOfElements do
      begin
      FEL_Nos[i] := 0;
      FEL_Qty[i] := 0;
      end;
end;

function TSpecies.CalcMolMass : double;
var i : integer;
begin
  result := 0;
  for i := 1 to MaxNumberOfElements do
    if FEL_Nos[i] <> 104 then
      result := result + AtomicData[FEL_Nos[i]].Atomic_Mass * FEL_Qty[i];
  result := result - FIonCharge * AtomicData[104].Atomic_Mass;  // e(-g)
end;

function TSpecies.Cp(T: single) : single;
var i : integer;
    X : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
      Result := (F[2]+2.0*(((2.0*F[7]*X+F[6])*3.0*X+
          F[5])*X+F[3]/(X*X)));   {J/mol}
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

function TSpecies.S(T: single) : single;
var i : integer;
    X : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
      Result := (F[1]+F[2]*(1+LN(X))-F[3]/SQR(X)+
          ((4.0*F[7]*X+3.0*F[6])*X+2.0*F[5])*X); {J/mol}
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

function TSpecies.Hfull(T: single) : single;
var i : integer;
    X : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
      Result := ((((3.0*F[7]*X+2.0*F[6])*X+F[5])*X+
          F[2])*X-F[4]-2.0*F[3]/X)*10; {kJ/mol}
      Result := Result + (DH298 - H298H0);
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

function TSpecies.H(T: single) : single;
var i : integer;
    X : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
      Result := ((((3.0*F[7]*X+2.0*F[6])*X+F[5])*X+
          F[2])*X-F[4]-2.0*F[3]/X)*10; {kJ/mol}
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

function TSpecies.G(T: single) : single;
var i : integer;
    X,H,S : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
        begin
        H := ((((3.0*F[7]*X+2.0*F[6])*X+F[5])*X+
              F[2])*X-F[4]-2.0*F[3]/X)*10000; {J/mol}     // Hfull
        H := H  + (DH298 - H298H0)*1000;
        S := (F[1]+F[2]+F[2]*LN(X)-F[3]/SQR(X)+
             ((4.0*F[7]*X+3.0*F[6])*X+2.0*F[5])*X); {J/mol}
        end;
      Result := (H - T*S)/1000;    {kJ/mol}
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

function TSpecies.F(T: single) : single;
var i : integer;
    X,H,S : single;
begin
  for i := 0 to FPolynoms.Count-1 do
    begin
    Polynom := FPolynoms.GetPolyn(i);
    if (Polynom.Tmin <= T) and (T <= Polynom.Tmax) then
      begin
      X := T*1.e-4;
      with Polynom do
        begin
        H := ((((3.0*F[7]*X+2.0*F[6])*X+F[5])*X+
              F[2])*X-F[4]-2.0*F[3]/X)*10000; {J/mol}
        S := (F[1]+F[2]+F[2]*LN(X)-F[3]/SQR(X)+
             ((4.0*F[7]*X+3.0*F[6])*X+2.0*F[5])*X); {J/mol}
        end;
      Result := S - H/T;      {J/mol}
      Exit;
      end;
    end;
  Raise Exception.Create('Out of temperature range: '+FloatToStr(T));
end;

procedure TSpecies.AddExPolynoms;
var Cp_,H_,S_ : single;
begin
  if not Assigned(FPolynoms) then Raise Exception.Create('No polynoms available');
// add higher polynom
  Cp_ := Cp(FPolynoms.Tmax);
  H_ := H(FPolynoms.Tmax)/10;
  S_ := S(FPolynoms.Tmax);
  FPolynoms.AddPol(FPolynoms.Tmax, 1.e30,
                   [S_ - Cp_*(Ln(FPolynoms.Tmax/10000) + 1),
                    Cp_, 0,
                    Cp_*FPolynoms.Tmax/10000 - H_,
                    0,0,0]);
// add Lower polynom
  if FPolynoms.Tmin > 1 then
    begin
    Cp_ := Cp(FPolynoms.Tmin);
    H_ := H(FPolynoms.Tmin)/10;
    S_ := S(FPolynoms.Tmin);
    FPolynoms.AddPol(1, FPolynoms.Tmin,
                     [S_ - Cp_*(Ln(FPolynoms.Tmin/10000) + 1),
                      Cp_, 0,
                      Cp_*FPolynoms.Tmin/10000 - H_,
                      0,0,0]);
    end;
end;

procedure TSpecies.SetFormula(S : ST20);
begin
  FFormula := S;
end;

procedure TSpecies.SetState(S : ST20);
begin
  FState := S;
end;

procedure TSpecies.SetPrecClass(S : ST4);
begin
  FPrecClass := S;
end;

procedure TSpecies.SetName(S : ST50);
begin
  FName := S;
end;

procedure TSpecies.SetMolMass(S : double);
begin
  FMolMass := S;
end;

procedure TSpecies.SetTref(S : single);
begin
  FTref := S;
end;

procedure TSpecies.SetReaction(S : ST50);
begin
  FReaction := S;
end;

procedure TSpecies.SetDHR(S : single);
begin
  FDHR := S;
end;

procedure TSpecies.SetDH0(S : single);
begin
  FDH0 := S;
end;

procedure TSpecies.SetDH298(S : single);
begin
  FDH298 := S;
end;

procedure TSpecies.SetCp298(S : single);
begin
  FCp298 := S;
end;

procedure TSpecies.SetS298(S : single);
begin
  FS298 := S;
end;

procedure TSpecies.SetH298H0(S : single);
begin
  FH298H0 := S;
end;

procedure TSpecies.SetSnucl(S : single);
begin
  FSnucl := S;
end;

procedure TSpecies.SetPrefUnit(S : ST3);
begin
  FPrefUnit := S;
end;

procedure TSpecies.SetDate(S : ST8);
begin
  FDate := S;
end;

procedure TSpecies.SetNelts(S : byte);
begin
  FNelts := S;
end;

procedure TSpecies.SetNpols(S : byte);
begin
  FNPols := S;
end;

procedure TSpecies.SetPstate(S : byte);
begin
  FPstate := S;
end;

procedure TSpecies.SetAddrPol(S : integer);
begin
  FAddrPol := S;
end;

procedure TSpecies.SetEL_Nos(i: integer; S : TNos);
begin
  if (i > High (FEL_Nos)) or (i < Low(FEL_Nos)) then
    raise Exception.Create('El_Nos index out of range: '+inttostr(i));
  FEL_Nos[i] := S;
end;

procedure TSpecies.SetEL_Qty(i: integer; S : TQty);
begin
  if (i > High (FEL_Qty)) or (i < Low(FEL_Qty)) then
    raise Exception.Create('El_Qty index out of range: '+inttostr(i));
  FEL_Qty[i] := S;
end;

function TSpecies.GetFEL_Nos(i: integer) : TNos;
begin
  if (i > High (FEL_Nos)) or (i < Low(FEL_Nos)) then
    raise Exception.Create('El_Nos index out of range: '+inttostr(i));
  Result := FEL_Nos[i];
end;

function TSpecies.GetFEL_Qty(i: integer) : TQty;
begin
  if (i > High (FEL_Qty)) or (i < Low(FEL_Qty)) then
    raise Exception.Create('El_Qty index out of range: '+inttostr(i));
  Result := FEL_Qty[i];
end;

procedure TSpecies.SetPolynoms(S: TPolynoms);
begin
  FPolynoms := S;
end;

procedure TOneSpecies.SetFormula(S: ST20);
begin
  FFormula := S;
end;

procedure TOneSpecies.SetIndex(S: integer);
begin
  Findex := S;
end;

procedure TOneSpecies.SetNos(S: TEL_Nos);
begin
  FNos := S;
end;

constructor TListOfSpecies.Create;
begin
  inherited Create;
  FList := TStringList.Create;
  FList.Duplicates := dupAccept;
  FList.Sorted := True;
end;

destructor TListOfSpecies.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

procedure TListOfSpecies.AddSubstance(S: TOneSpecies);
begin
  FList.AddObject(S.Formula,S);
end;

procedure TListOfSpecies.Clear;
begin
  FList.Clear;
end;

function TListOfSpecies.SubstanceByName(pIndex: ST20) : TOneSpecies;
var i : integer;
begin
  i := FList.IndexOf(pIndex);
  if i >= 0 then
    begin
    Result := FList.Objects[i] as TOneSpecies;
    Exit;
    end;
    Raise Exception.Create('Substance '+pIndex+' not found');
end;

function TListOfSpecies.SubstanceByIndex(pIndex: integer) : TOneSpecies;
begin
    if (pIndex < 0) or (pIndex > FList.Count-1) then
      raise Exception.Create('Invalid index: '+IntToStr(pIndex));
    Result := FList.Objects[pIndex] as TOneSpecies;
end;

// TPXListOfSpecies

constructor TTPXListOfSpecies.Create;
begin
  inherited Create;
  FList := TStringList.Create;
  FList.Duplicates := dupAccept; //dupError;
  FList.Sorted := True;
end;

destructor TTPXListOfSpecies.Destroy;
begin
  FList.Free;
  inherited Destroy;
end;

procedure TTPXListOfSpecies.AddSubstance(TPX_KOD: st30; S: TOneSpecies);
begin
  FList.AddObject(TPX_KOD,S);
end;

procedure TTPXListOfSpecies.Clear;
begin
  FList.Clear;
end;

function TTPXListOfSpecies.GetFCount : integer;
begin
  Result := FList.Count;
end;

function TTPXListOfSpecies.SubstanceByName(pIndex: ST20) : TOneSpecies;
var i : integer;
begin
  for i := 0 to FList.Count-1 do
    begin
    OneSpecies := FList.Objects[i] as TOneSpecies;
    if OneSpecies.Formula = pIndex then
      begin
      Result := OneSpecies;
      Exit;
      end;
    end;
    Raise Exception.Create('Substance '+pIndex+' not found');
end;

function TTPXListOfSpecies.SubstanceByIndex(pIndex: integer) : TOneSpecies;
begin
    if (pIndex < 0) or (pIndex > FList.Count-1) then
      raise Exception.Create('Invalid index: '+IntToStr(pIndex));
    Result := FList.Objects[pIndex] as TOneSpecies;
end;

//  TFullList
constructor TFullList.Create;
begin
  inherited Create;
  FList := TStringList.Create;
end;

destructor TFullList.Destroy;
var i : integer;
begin
  for i := 0 to FList.Count-1 do
    begin
    Species := FList.Objects[i] as TSpecies;
    Species.Free;
    end;
  FList.Free;
  inherited Destroy;
end;

procedure TFullList.Clear;
var i : integer;
begin
  for i := 0 to FList.Count-1 do
    begin
    Species := FList.Objects[i] as TSpecies;
    Species.Free;
    end;
  FList.Clear;
end;

procedure TFullList.AddSubstance(S: TSpecies);
begin
  FList.AddObject(S.Formula,S);
end;

function TFullList.GetFCount : integer;
begin
  Result := FList.Count;
end;

function TFullList.SubstanceByName(pIndex: ST20) : TSpecies;
var i : integer;
begin
  if pIndex < ' ' then
    Raise Exception.Create('Formula is not specified');
  i := FList.IndexOf(pIndex);
  if i >= 0 then
    begin
    Result := FList.Objects[i] as TSpecies;
    Exit;
    end;
  Raise Exception.Create('Substance '+pIndex+' not found');
end;

function TFullList.SubstanceByIndex(pIndex: integer) : TSpecies;
begin
    if (pIndex < 0) or (pIndex > FList.Count-1) then
      raise Exception.Create('Invalid index: '+IntToStr(pIndex));
    Result := FList.Objects[pIndex] as TSpecies;
end;

end.